#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Provides data access classes for accessing and managing MythTV data
"""

from MythStatic import *
from MythBase import *

import re, sys, socket, os
from time import mktime, strftime
from socket import gethostbyaddr, gethostname

def ftopen(file, type, forceremote=False):

    def _openRemote(host, filename, sgroup, type):
        if isinstance(sgroup, StorageGroup):
            sgroup = sgroup.groupname
        return FileTransfer(host, filename, sgroup, type)

    def _openLocal(host, filename, sgroup, type):
        if not isinstance(sgroup, StorageGroup):
            sgroup = self.db.getStorageGroup(groupname=sgroup, hostname=host)[0]
        return open(sgroup.dirname+filename, type)

    db = MythDBConn()
    reuri = re.compile(\
        'myth://((?P<group>.*)@)?(?P<host>[a-zA-Z0-9\.]*)(:[0-9]*)?/(?P<file>.*)')
    reip = re.compile('(?:\d{1,3}\.){3}\d{1,3}')

    if isinstance(file, Program):
        # process Program object
        file = file.filename
    elif isinstance(file, Recorded):
        # process Recorded object
        file = (file.hostname, file.basename, file.storagegroup)

    if isinstance(file, tuple) or isinstance(file, list):
        # process tuple (host, filename, storagegroup)
        if len(file) != 3:
            #log.Msg(CRITICAL, 'Incorrect FileTransfer() input size')
            raise MythError('Incorrect FileTransfer input tuple size: %d'\
                    % len(file))
        (host, filename, sgroup) = file
    elif isinstance(file, str):
        # process URI (myth://<group>@<host>[:<port>]/<path/to/file>)
        match = reuri.match(file)
        if match is None:
            #log.Msg(...)
            raise MythError('Invalid FileTransfer input string: '+file)
        host = match.group('host')
        filename = match.group('file')
        sgroup = match.group('group')
        if sgroup is None:
            sgroup = 'Default'
    else:
        #log.Msg(...)
        raise MythError('Invalid FileTransfer input type; %s' %type(file))

    # remote leading slash if exists
    filename = filename.lstrip('/')

    # get full system name
    if reip.match(host):
        host = gethostbyaddr(host)[0]

    # user forced to remote access
    if forceremote:
        return _openRemote(host, filename, sgroup, type)

    if isinstance(sgroup, StorageGroup):
        sgs = [sgroup]
    else:
        sgs = db.getStorageGroup(groupname=sgroup, hostname=host)

    if type == 'w':
        # force remote, let mythtv decide how to disk balance
        if (len(sgs) > 1):
            return _openRemote(host, filename, sgroup, type)
        # group not local, use remote access
        if not sgs[0].local:
            return _openRemote(host, filename, sgs[0], type)
        # open local file
        return _openLocal(host, filename, sgs[0], type)
    else:
        for sg in sgs:
            if sg.local:
                if os.access(sg.dirname+filename, os.F_OK):
                    # file found, open local
                    return _openLocal(host, filename, sg, type)
        # file not found, open remote
        return _openRemote(host, filename, sgroup, type)

class FileTransfer( MythBEConn ):
    """
    A connection to mythbackend intended for file transfers
    """

    def __init__(self, host, filename, sgroup, type):
        if type not in ('r','w'):
            raise MythError("FileTransfer type must be read ('r') "+
                                "or write ('w')")

        self.host = host
        self.filename = filename
        self.sgroup = sgroup
        self.tsize = 2**15
        self.open = False
        # create control socket
        self.control = MythBEConn(host,'Playback')
        # continue normal Backend initialization
        MythBEConn.__init__(self,host,type)
        self.open = True

    def announce(self, type):
        # replacement announce for Backend object
        if type == 'w':
            self.w = True
        elif type == 'r':
            self.w = False
        res = self.backendCommand('ANN FileTransfer %s %d %d %s' \
                    % (socket.gethostname(), self.w, False,
                        BACKEND_SEP.join(['-1',self.filename,self.sgroup])))
        if res.split(BACKEND_SEP)[0] != 'OK':
            #log.Msg(CRITICAL, 'Unexpected answer to ANN command: %s', res)
            raise MythError('Unexpected answer to ANN command: %s' % res)
        else:
            #log.Msg(INFO, 'Successfully connected mythbackend at %s:%d',
            #            self.host, self.port)
            sp = res.split(BACKEND_SEP)
            self.sockno = int(sp[1])
            self.pos = 0
            self.size = (int(sp[2]) + (int(sp[3])<0))*2**32 + int(sp[3])

    def __del__(self):
        if not self.open:
            return
        self.control.backendCommand('QUERY_FILETRANSFER '\
                    +BACKEND_SEP.join([str(self.sockno), 'JOIN']))
        self.socket.shutdown(1)
        self.socket.close()
        self.open = False

    def tell(self):
        """
        Return the current offset from the beginning of the file
        """
        return self.pos

    def close(self):
        """
        Close the data transfer socket
        """
        self.__del__()

    def rewind(self):
        """
        Seek back to the start of the file
        """
        self.seek(0)

    def read(self, size):
        """
        Read a block of data, requests over 64KB will be buffered internally
        """
        if self.w:
            print 'Error: attempting to read from a write-only socket'
            return
        if size == 0:
            return ''
        if size > self.size - self.pos:
            size = self.size - self.pos
        csize = size
        rsize = 0
        if csize > self.tsize:
            csize = self.tsize
            rsize = size - csize

        res = self.control.backendCommand('QUERY_FILETRANSFER '+BACKEND_SEP\
                    .join([str(self.sockno),'REQUEST_BLOCK',str(csize)]))
        self.pos += int(res)
#       if int(res) == csize:
#           if csize < size:
#               self.tsize += 8192
#       else:
#           self.tsize -= 8192
#           rsize = size - int(res)
#       print 'resizing buffer to %d' % self.tsize

        return self.socket.recv(int(res)) + self.read(rsize)

    def write(self, data):
        """
        Write a block of data, requests over 64KB will be buffered internally
        """
        if not self.w:
            print 'Error: attempting to write to a read-only socket'
            return
        size = len(data)
        buff = ''
        if size == 0:
            return
        if size > self.tsize:
            size = self.tsize
            buff = data[size:]
            data = data[:size]
        self.pos += int(self.socket.send(data))
        self.control.backendCommand('QUERY_FILETRANSFER '+BACKEND_SEP\
                    .join([str(self.sockno),'WRITE_BLOCK',str(size)]))
        self.write(buff)
        return


    def seek(self, offset, whence=0):
        """
        Seek 'offset' number of bytes
           whence==0 - from start of file
                   whence==1 - from current position
        """
        if whence == 0:
            if offset < 0:
                offset = 0
            if offset > self.size:
                offset = self.size
        elif whence == 1:
            if offset + self.pos < 0:
                offset = -self.pos
            if offset + self.pos > self.size:
                offset = self.size - self.pos
        elif whence == 2:
            if offset > 0:
                offset = 0
            if offset < -self.size:
                offset = -self.size
        #else:
            #log.Msg(CRITICAL, 'Whence can only be 0, 1, or 2')

        curhigh,curlow = self.splitInt(self.pos)
        offhigh,offlow = self.splitInt(offset)

        res = self.control.backendCommand('QUERY_FILETRANSFER '+BACKEND_SEP\
                        .join([str(self.sockno),'SEEK',str(offhigh),
                                str(offlow),str(whence),str(curhigh),
                                str(curlow)])\
                    ).split(BACKEND_SEP)
        self.pos = self.joinInt(res[0],res[1])


class FreeSpace(DictData):
    """
    Represents a Storage Group entry
    """
    field_order = [ 'host',         'path',     'islocal',
                    'disknumber',   'sgroupid', 'blocksize',
                    'ts_high',      'ts_low',   'us_high',
                    'us_low']
    field_type = [3, 3, 2, 0, 0, 0, 0, 0, 0, 0]
    def __str__(self):
        return "Storage Group %d (%s@%s)"\
                    % (self.sgroupid, self.path, self.host)

    def __repr__(self):
        return "Storage Group %d (%s@%s)"\
                    % (self.sgroupid, self.path, self.host)

    def __init__(self, data):
        DictData.__init__(self, data)
        self.totalspace = self.joinInt(self.ts_high, self.ts_low)
        self.usedspace = self.joinInt(self.us_high, self.us_low)
        self.freespace = self.totalspace - self.usedspace


class Program(DictData):
    """
    Represents a program with all the detail known.
    """

    # field names and type in order
    #   0: means integer
    #   1: means float
    #   2: means bool
    #   3: means string
    #   4: means date
    field_order = [ 'title',        'subtitle',     'description',
                    'category',     'chanid',       'channum',
                    'callsign',     'channame',     'filename',
                    'fs_high',      'fs_low',       'starttime',
                    'endtime',      'duplicate',    'shareable',
                    'findid',       'hostname',     'sourceid',
                    'cardid',       'inputid',      'recpriority',
                    'recstatus',    'recordid',     'rectype',
                    'dupin',        'dupmethod',    'recstartts',
                    'recendts',     'repeat',       'programflags',
                    'recgroup',     'commfree',     'outputfilters',
                    'seriesid',     'programid',    'lastmodified',
                    'stars',        'airdate',      'hasairdate',
                    'playgroup',    'recpriority2', 'parentid',
                    'storagegroup', 'audio_props',  'video_props',
                    'subtitle_type','year']
    field_type = [  3,      3,      3,
                    3,      0,      3,
                    3,      3,      3,
                    0,      0,      4,
                    4,      0,      0,
                    0,      3,      0,
                    0,      0,      0,
                    0,      0,      3,
                    0,      0,      4,
                    4,      0,      3,
                    3,      0,      3,
                    3,      3,      3,
                    1,      3,      0,
                    3,      0,      3,
                    3,      0,      0,
                    0,      0]
    def __str__(self):
        return u"%s (%s)" % (self.title,
                    self.starttime.strftime('%Y-%m-%d %H:%M:%S'))

    def __repr__(self):
        return u"%s (%s)".encode('utf-8') % (self.title,
                    self.starttime.strftime('%Y-%m-%d %H:%M:%S'))

    def __init__(self, data):
        DictData.__init__(self, data)
        self.filesize = self.joinInt(self.fs_high,self.fs_low)

    def toString(self):
        data = []
        for i in range(0,PROGRAM_FIELDS):
            if self.data[self.field_order[i]] == None:
                datum = ''
            elif self.field_type[i] in (0,1,3):
                datum = str(self.data[self.field_order[i]])
            elif self.field_type[i] == 2:
                datum = str(int(self.data[self.field_order[i]]))
            elif self.field_type[i] == 4:
                datum = str(int(mktime(\
                        self.data[self.field_order[i]].timetuple())))
            data.append(datum)
        return BACKEND_SEP.join(data)

    def getRecorded(self):
        """
        Return a Recorded object for manipulation of recording
        """
        return Recorded((self.chanid,self.recstartts))

    def open(self, type='r'):
        if type != 'r':
            MythError('Program only supports read-only access to files')
        return ftopen(self, 'r')


class Recorded( DBDataWrite ):
    """
    Class for managing recordings
    """
    table = 'recorded'
    where = 'chanid=%s AND starttime=%s'

    class _Cast( DBDataCRef ):
        table = 'people'
        ref = 'recordedcredits'
        t_field = 'person'
        t_add = ['name']
        r_field = 'person'
        r_add = ['role']
        w_field = ['chanid','starttime']
        def __repr__(self):
            if len(self.data) == 0:
                return 'No cast'
            cast = []
            for cref in self.field_order:
                cast.append(self.data[cref].name)
            return ', '.join(cast)

    def __str__(self):
        return u"%s (%s)" % (self.title,
                    self.starttime.strftime('%Y-%m-%d %H:%M:%S'))

    def __repr__(self):
        return u"%s (%s)".encode('utf-8') % (self.title,
                    self.starttime.strftime('%Y-%m-%d %H:%M:%S'))

    def __init__(self, data=None, db=None):
        """
        Creates a new recording object.
        Should be supplied with database if possible.
        Can be initialized in three manners:
            data = (int(chanid),int(starttime)):
                    initialize object to existing entry in recorded
            data = list:
                   use data from a row returned from 'select * from recorded'
            data = dictionary:
                    initialize object to new recording, using dictionary
        """
        self._setDefs()
        self.db = MythDBConn(db)
        self.field_order = self.db.tablefields[self.table]

        if isinstance(data,tuple) or isinstance(data,list):
            if len(data) == 2:
                self.wheredat = tuple(data)
                self._pull()
            elif len(data) == len(self.field_order):
                self._import(data)
                self.wheredat = (self.chanid,self.starttime)
            else:
                raise MythError('Incorrect list input length to Recorded()')
        elif isinstance(data,dict):
            self.create(data)
            self.wheredat = (data.chanid,data.starttime)
            self._pull()
        else:
            raise MythError('Invalid input type to Recorded(): '+str(type(data)))
        if self.data is None:
            raise MythError('Failed to read Recorded() data:'+str(self.wheredat))
        self.cast = self._Cast(self.wheredat,self.db)

    def delete(self, rerecord=False):
        if rerecord:
            c = self.db.cursor()
            c.execute("""UPDATE %s SET duplicate=0 WHERE %s"""\
                % ('oldrecorded',self.where), self.wheredat)
            c.close()
        # call for deletion over backend socket

    def open(self, type='r'):
        return ftopen(self, type)


class Job( DBDataWrite ):
    """
    Class for managing tasks within mythtv's jobqueue
    """
    # types
    NONE         = 0x0000
    SYSTEMJOB    = 0x00ff
    TRANSCODE    = 0x0001
    COMMFLAG     = 0x0002
    USERJOB      = 0xff00
    USERJOB1     = 0x0100
    USERJOB2     = 0x0200
    USERJOB3     = 0x0300
    USERJOB4     = 0x0400
    # cmds
    RUN          = 0x0000
    PAUSE        = 0x0001
    RESUME       = 0x0002
    STOP         = 0x0004
    RESTART      = 0x0008
    # flags
    NO_FLAGS     = 0x0000
    USE_CUTLIST  = 0x0001
    LIVE_REC     = 0x0002
    EXTERNAL     = 0x0004
    # statuses
    UNKNOWN      = 0x0000
    QUEUED       = 0x0001
    PENDING      = 0x0002
    STARTING     = 0x0003
    RUNNING      = 0x0004
    STOPPING     = 0x0005
    PAUSED       = 0x0006
    RETRY        = 0x0007
    ERRORING     = 0x0008
    ABORTING     = 0x0008
    DONE         = 0x0100
    FINISHED     = 0x0110
    ABORTED      = 0x0120
    ERRORED      = 0x0130
    CANCELLED    = 0x0140

    table = 'jobqueue'
    def __init__(self, data=None, db=None):
        """
        Creates a new job object.
        Should be supplied with database if possible.
        Can be initialized in three manners:
            data = int(jobid) ---or--- data = (int(chanid),int(starttime)):
                    initialize object to existing entry in jobqueue
            data = dictionary:
                    initialize object to new job, using data from dictionary
        """
        self._setDefs()
        self.db = MythDBConn(db)
        self.field_order = self.db.tablefields[self.table]

        if isinstance(data,int):
            self.where = 'id=%s'
            self.wheredat = (data,)
        elif isinstance(data,tuple) or isinstance(data,list):
            if len(data) != 2:
                raise MythError('Incorrect list input length to Job()')
            self.where = 'chanid=%s AND starttime=%s'
            self.wheredat = tuple(data)
        elif isinstance(data,dict):
            self.where = 'id=%s'
            self.wheredat = (self.create(data),)
        else:
            raise MythError('Invalid input type to Job(): '+str(type(data)))
        self._getData()
        if self.data is None:
            raise MythError('Failed to read Job() data: ' + str(self.wheredat))

    def setComment(self,comment):
        self.comment = comment
        self.update()

    def setStatus(self,status):
        self.status = status
        self.update()

class Channel(DBDataWrite):
    """
    Represents a single channel from the channel table
    """
    table = 'channel'
    where = 'chanid=%s'
    
    def __str__(self):
        return u"%s (%s)" % (self.chanid, self.name)

    def __repr__(self):
        return u"%s (%s)".encode('utf-8') % (self.chanid, self.name)

    def __init__(self,data,db=None):
        """
        Creates a new channel object.
        Should be supplied with database if possible.
        Can be initialized in three manners:
            data = chanid:
                    initialize object to existing entry in recorded
            data = list/tuple:
                   use data from a row returned from 'select * from channel'
            data = dictionary:
                    initialize object to new recording, using dictionary
        """
        self._setDefs()
        self.db = MythDBConn(db)
        self.field_order = self.db.tablefields[self.table]

        if isinstance(data, int) or isinstance(data, str):
            self.wheredat = (int(data))
            self._pull()
        elif isinstance(data,tuple) or isinstance(data,list):
            if len(data) == len(self.field_order):
                self._import(data)
                self.wheredat = (self.chanid)
            else:
                raise MythError('Incorrect list input length to Recorded()')
        elif isinstance(data,dict):
            self.create(data)
            self.wheredat = (data.chanid)
            self._pull()
        else:
            raise MythError('Invalid input type to Channel(): '+str(type(data)))
        if self.data is None:
            raise MythError('Failed to read Channel() data:'+str(self.wheredat))
    
#    def openIcon(self, mode='r'):
    

class Guide(DBData):
    """
    Represents a single program from the program guide
    """
    table = 'program'
    where = 'chanid=%s AND starttime=%s'
    
    def __str__(self):
        return u"%s (%s)" % (self.title,
                self.starttime.strftime('%Y-%m-%d %H:%M:%S'))

    def __repr__(self):
        return u"%s (%s)".encode('utf-8') % (self.title,
                self.starttime.strftime('%Y-%m-%d %H:%M:%S'))

    def __init__(self,data,db=None):
        DBData.__init__(self,data,db)
        self.wheredat = (self.chanid,self.starttime)



class Video( DBDataWrite ):
    """
    Class for managing and creating metadata for a single video file
    """
    field_type = 'Pass'
    table = 'videometadata'
    where = 'intid=%s'
    defaults = {'subtitle':u'',             'director':u'Unknown',
                'rating':u'NR',             'inetref':u'00000000',
                'year':1895,                'userrating':0.0,
                'length':0,                 'showlevel':1,
                'coverfile':u'No Cover',    'host':u'',
                'intid':None}

    def __init__(self, id=None, data=None, db=None):
        """
        Creates a new video object.
        Should be supplied with existing database object if possible.
        Can be initialized in three manners, or left uninitialized.
            id = intid, data = None:
                    initialize object to an existing video entry
            id = intid, data = dictionary:
                    initialize object to an existing video entry
                    and update with info in dictionary
            id = none, data = list:
                    initialize object to an existing video entry
                    given by a row returned from 'select * from videometadata'
            id = none, data = dictionary:
                    initialize object to a new video entry
                    using info from dictionary
            id = none, data = none:
                leave object empty
        """
        self._setDefs()
        self.db = MythDBConn(db)
        self.field_order = self.db.tablefields[self.table]
        
        if id:
            self.wheredat = (id,)
            self._pull()
            if len(self.data) == 0:
                raise MythError(\
                        'Invalid intid used during initialization of Video()')
            if isinstance(data,dict):
                self._import(data)
                self.update()
        elif isinstance(data,list) or isinstance(data,tuple):
            self._import(data)
            self.wheredat = (self.intid,)
        elif isinstance(data,dict):
            self.wheredat = (self.create(data),)
            self._pull()
        self.cast = self._Cast((self.intid,), self.db)
        self.genre = self._Genre((self.intid,), self.db)
        self.country = self._Country((self.intid,), self.db)

    def __repr__(self):
        return self.__str__().encode('utf-8')

    def __str__(self):
        res = self.title
        if self.season and self.episode:
            res += u' - %dx%02d' % (self.season, self.episode)
        if self.subtitle:
            res += u' - '+self.subtitle
        return res

    class _Cast( DBDataCRef ):
        table = 'videocast'
        ref = 'videometadatacast'
        t_field = 'intid'
        t_add = ['cast']
        r_field = 'idcast'
        r_add = []
        w_field = ['idvideo']
        def __repr__(self):
            if len(self.data) == 0:
                return ''
            cast = []
            for cref in self.field_order:
                cast.append(self.data[cref].cast)
            return ', '.join(cast)

    class _Genre( DBDataCRef ):
        table = 'videogenre'
        ref = 'videometadatagenre'
        t_field = 'intid'
        t_add = ['genre']
        r_field = 'idgenre'
        r_add = []
        w_field = ['idvideo']
        def __repr__(self):
            if len(self.data) == 0:
                return ''
            cast = []
            for cref in self.field_order:
                cast.append(self.data[cref].genre)
            return ', '.join(cast)

    class _Country( DBDataCRef ):
        table = 'videocountry'
        ref = 'videometadatacountry'
        t_field = 'intid'
        t_add = ['country']
        r_field = 'idcountry'
        r_add = []
        w_field = ['idvideo']
        def __repr__(self):
            if len(self.data) == 0:
                return ''
            cast = []
            for cref in self.field_order:
                cast.append(self.data[cref].country)
            return ', '.join(cast)

    def _open(self, type, mode='r'):
        """
        Open data socket
        """
        sgroup = {  'filename':'Videos',        'banner':'Banners',
                    'coverfile':'Coverart',     'fanart':'Fanart',
                    'screenshot':'Screenshots', 'trailer':'Trailers'}
        if self.data is None:
            return None
        if type not in sgroup:
            raise MythError('Video._open() - '+\
                            'Trying to open invalid file type: '+str(type))
        SG = self.db.getStorageGroup(sgroup[type], self.host)
        if len(SG) == 0:
            SG = self.db.getStorageGroup('Videos', self.host)
            if len(SG) == 0:
                raise MythError('Could not find storagegroup')
        return ftopen((self.host,self.data[type],SG[0].groupname),mode)

    def delete(self):
        """
        Delete video entry from database
        """
        if self.data is None:
            return
        self.cleanCast()
        self.cleanGenre()
        self.cleanCountry()
        DBDataWrite.delete(self)

    def open(self,mode='r'):
        """Open file pointer to video"""
        return self._open('filename',mode)

    def openBanner(self,mode='r'):
        """Open file pointer to banner image"""
        return self._open('banner',mode)

    def openCoverart(self,mode='r'):
        """Open file pointer to coverart image"""
        return self._open('coverfile',mode)

    def openFanart(self,mode='r'):
        """Open file pointer to fanart image"""
        return self._open('fanart',mode)

    def openScreenshot(self,mode='r'):
        """Open file pointer to screenshot image"""
        return self._open('screenshot',mode)

    def openTrailer(self,mode='r'):
        """Open file pointer to trailer"""
        return self._open('trailer',mode)





# -*- coding: utf-8 -*-
"""
Contains any static and global variables for MythTV Python Bindings
"""

MYTHXMLCONN = None
MYTHDBCONN = {}
SCHEMA_VERSION = 1246
MVSCHEMA_VERSION = 1031
PROTO_VERSION = 51
PROGRAM_FIELDS = 47
BACKEND_SEP = '[]:[]'


#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Provides base classes for accessing MythTV
"""

from MythStatic import *

import os, re, socket, MySQLdb
import xml.dom.minidom as minidom
from datetime import datetime
from urllib import urlopen


class DictData( object ):
    """
    Basic class for providing management of data resembling dictionaries.
    Must be subclassed for use.
    """
    # Requires two lists of equal length to describe input fields
    #   The first list is one of field names
    #   The second list is one of field types
    #     0: integer
    #     1: float
    #     2: bool
    #     3: string
    #     4: date -- processed by datetime.fromtimestamp(int(value))
    #   The second list can be a single string 'Pass'
    #               to pass all values untouched
    
    #requires:
    #  field_order
    #  field_type

    ### class emulation functions
    def __getattr__(self,name):
        # function for class attribute access of data fields
        # accesses real attributes, falls back to data fields, and errors
        if name in self.__dict__:
            return self.__dict__[name]
        elif name in self.field_order:
            return self.data[name]
        else:
            raise AttributeError("'DictData' object has no attribute '%s'"%name)

    def __setattr__(self,name,value):
        # function for class attribute access of data fields
        # sets value in data fields, falling back to real attributes
        if name in self.field_order:
            self.data[name] = value
        else:
            self.__dict__[name] = value

    ### container emulation functions
    def __getitem__(self,key):
        # function for list-like and dist-like access to content
        # as a side effect of dual operation, dict data cannot be indexed
        #       keyed off integer values
        if isinstance(key,str):
            if key in self.data:
                return self.data[key]
            else:
                raise KeyError("'DictData' object has no key '%s'" %key)
        elif isinstance(key,int):
            if key in range(0,len(self.field_order)):
                if self.field_order[key] in self.data:
                    return self.data[self.field_order[key]]
            else:
                raise IndexError("Index out of bounds")
        raise TypeError('Data access must use string or integer')

    def __setitem__(self,key,value):
        # function for list-like and dist-like access to content
        # does not allow setting of data not already in field_order
        if isinstance(key,str):
            if key in self.field_order:
                self.data[key] = value
            else:
                raise KeyError("'DictData' does not allow new data")
        elif isinstance(key,int):
            if key in range(0,len(self.field_order)):
                self.data[self.field_order[key]] = value
            else:
                raise IndexError("'DictData' does not allow new data")
        else:
            raise KeyError('Value not usable in DictData')
            
    def __contains__(self,item):
        return bool(item in self.data.keys())

    def __iter__(self):
        self.field = 0
        self.itermode = 0
        return self
        
    def iterkeys(self):
        self.field = 0
        self.itermode = 1
        return self
        
    def itervalues(self):
        self.field = 0
        self.itermode = 0
        return self
    
    def iteriterms(self):
        self.field = 0
        self.itermode = 2
        return self

    def next(self):
        if self.field == len(self.field_order):
            del self.field
            del self.itermode
            raise StopIteration
        self.field += 1
        if self.itermode == 0:
            return self.data[self.field_order[self.field-1]]
        elif self.itermode == 1:
            return self.field_order[self.field-1]
        elif self.itermode == 2:
            return (self.field_order[self.field-1],
                    self.data[self.field_order[self.field-1]])
        
    def keys(self):
        return self.field_order
        
    def has_key(self,item):
        return self.__contains__(item)
        
    def values(self):
        return list(self)
        
    def get(self,key):
        return self.data[key]
        
    def items(self):
        pairs = []
        for field in self.field_order:
            pairs.append((field, self.data[field]))
        return tuple(pairs)
        
    def pop(self,key):
        raise NotImplementedError
    
    def popitem(self):
        raise NotImplementedError
        
    def update(self, *args, **keywords):
        # accepts database, iterable, or keywords to update the data
        if self.data is None: self.data = {}
        self.data.update(apply(self._distill, args, keywords))

    ### additional management functions
    def _setDefs(self):
        self.data = {}
        self.field = 0
        
    def __init__(self, data):
        self._setDefs()
        self.data.update(self._distill(self._process(data)))

    def _process(self, data):
        data = list(data)
        for i in range(0, len(data)):
            if data[i] == '':
                data[i] = None
            elif self.field_type == 'Pass':
                data[i] = data[i]
            elif self.field_type[i] == 0:
                data[i] = int(data[i])
            elif self.field_type[i] == 1:
                data[i] = float(data[i])
            elif self.field_type[i] == 2:
                data[i] = bool(data[i])
            elif self.field_type[i] == 3:
                data[i] = data[i]
            elif self.field_type[i] == 4:
                data[i] = datetime.fromtimestamp(int(data[i]))
        return tuple(data)

    def _distill(self, *args, **keywords):
        # accepts dict, iterable, or keywords and outputs a distilled dict
        data = {}
        for arg in args:
            if isinstance(arg,dict):
                for key in arg.keys():
                    data[key] = arg[key]
            elif isinstance(arg, list) or isinstance(arg, tuple):
                if len(arg) == len(self.field_order):
                    data.update(zip(self.field_order, arg))
                else:
                    for key,value in arg:
                        data[key] = value
        for key in keywords.keys():
            data[key] = keywords[key]
        for key in data:
            if key not in self.field_order:
                del data[key]
        return data

    def _import(self, *args, **keywords):
        if self.data is None: self.data = {}
        self.data.update(apply(self._distill, args, keywords))

    def joinInt(self,high,low):
        """
        Returns a single long from a pair of signed integers.
        """
        return (high + (low<0))*2**32 + low

    def listFields(self):
        """
        Returns a tuple of all available object fields.
        """
        return tuple(self.data.keys())

class DBData( DictData ):
    """
    Altered DictData adding several functions for dealing with individual rows.
    Must be subclassed for use.
    """
    field_type = 'Pass'

    def _setDefs(self):
        self.__dict__['field_order'] = []
        DictData._setDefs(self)

    def __init__(self,data=None,db=None):
        self._setDefs()
        self.db = MythDBConn(db)
        self.field_order = self.db.tablefields[self.table]
        
        if data is not None:
            self._import(self,data)

    def _pull(self):
        """Updates table with data pulled from database."""
        c = self.db.cursor()
        c.execute("""SELECT * FROM %s WHERE %s""" \
                % (self.table, self.where), self.wheredat)
        data = c.fetchone()
        c.close()
        if data is None:
            return
        self._import(data)

class DBDataWrite( DBData ):
    """
    Altered DBData, with support for writing back to the database.
    'defaults' dictionary allows default values when creating creating new
            and updating old entries
    Must be subclassed for use.
    """
    defaults = None
    def _sanitize(self,data):
        """Remove fields from dictionary that are not in database table."""
        for key in data:
            if key not in self.field_order:
                del data[key]
        if self.defaults is None:
            return data
        for key in self.defaults:
            if key in data:
                if self.defaults[key] is None:
                    del data[key]
                elif data[key] is None:
                    data[key] = self.defaults[key]
        return data

    def create(self,data):
        """
        Creates a new database entry using given dictionary.
        Only works on an uninitialized object.
        """
        if len(self.data) > 0:
            raise MythError('DBDataWrite object already bound to '+\
                                                    'existing instance')
        self._insert(self, data)
        data = self._sanitize(self.data)
        c = self.db.cursor()
        fields = ', '.join(data.keys())
        format_string = ', '.join(['%s' for d in data.values()])
        c.execute("""   INSERT INTO %s (%s)
                        VALUES(%s)"""\
                    % (self.table, fields, format_string), data.values())
        intid = c.lastrowid
        c.close()
        return intid

    def _push(self):
        """
        Updates database with data from local table.
        """
        if (self.where is None) or (self.wheredat is None):
            return
        c = self.db.cursor()
        data = self._sanitize(self.data)
        format_string = ', '.join(['%s = %%s' % d for d in data])
        sql_values = data.values()
        sql_values.extend(self.wheredat)

        c.execute("""   UPDATE %s SET %s WHERE %s"""\
                % (self.table, format_string, self.where), sql_values)
        c.close()
        self._pull()
        
    def update(self, *args, **keywords):
        apply(self._import, args, keywords)
        self._push()

    def delete(self):
        """
        Delete video entry from database.
        """
        if (self.where is None) or \
                (self.wheredat is None) or \
                (self.data is None):
            return
        c = self.db.cursor()
        c.execute("""   DELETE FROM %s WHERE %s"""\
                % (self.table, self.where), self.wheredat)
        c.close()

class DBDataCRef( DictData ):
    """
    Altered DictData for managing lists of crossreferenced data.
    """
    field_type = 'Pass'

    class SubData( DictData ):
        def _setDefs(self):
            self.__dict__['field_order'] = []
            DictData._setDefs(self)
        def __repr__(self):
            return str(tuple(self.data.values()))
        def __init__(self,data,fields):
            self._setDefs()
            self.field_order = fields
            self.field_type = 'Pass'
            self._import(data)

    def _setDefs(self):
        self.__dict__['field_order'] = []
        DictData._setDefs(self)
    def __repr__(self):
        return self.data.keys()

    def __init__(self,where,db=None):
        self._setDefs()
        self.db = MythDBConn(db)
        self.w_dat = tuple(where)
        
        if self.t_add is None: self.t_add = []
        if self.r_add is None: self.r_add = []
        
        tf = '%s.%s' % (self.table, self.t_field)
        rf = '%s.%s' % (self.ref, self.r_field)
        
        c = self.db.cursor()
        vals = ','.join([tf]+self.t_add+self.r_add)
        where = ' AND '.join('%s=%%s' % d for d in self.w_field)
        join = '%s=%s' % (tf,rf)
        c.execute("""SELECT %s FROM %s JOIN %s ON %s WHERE %s""" % \
                (vals, self.table, self.ref, join, where), self.w_dat)
        for entry in c.fetchall():
            self.field_order.append(entry[0])
            self.data[entry[0]] = self.SubData(entry[1:],self.t_add+self.r_add)

    def add(self,data):
        if len(data) != len(self.t_add+self.r_add):
            return

        # check for existing
        for dat in self.data:
            if data == dat.data.values():
                return

        c = self.db.cursor()
        
        # search for existing table data, add if needed
        tdata = data[:len(self.t_add)]
        where = ' AND '.join('%s=%%s' % d for d in self.t_add)
        count = c.execute("""SELECT %s FROM %s WHERE %s""" % \
                        (self.t_field, self.table, where), tdata)
        if count == 1:
            id = c.fetchone()[0]
        else:
            fields = ', '.join(self.t_add)
            format = ', '.join(['%s' for d in tdata])
            c.execute("""INSERT INTO %s (%s) VALUES(%s)"""\
                    % (self.table, fields, format), tdata)
            id = c.lastrowid
            
        # add cross reference
        rdata = data[-len(self.r_add):]+id+self.w_dat
        rfields = self.r_add+self.r_field+self.w_field
        fields = ', '.join(rfields)
        format = ', '.join(['%s' for d in rdata])
        c.execute("""INSERT INTO %s (%s) VALUES(%s)"""\
                    % (self.ref, fields, format), rdata)
        c.close()

        # add local entry
        self.field_order.append(id)
        self.data[id] = self.SubData(id,data)

    def delete(self,cref):
        if cref not in self.data:
            return

        c = self.db.cursor()
        wdata = cref+self.w_dat
        wfields = self.r_field+self.w_field
        where = ' AND '.join('%s=%ss' % d for d in wfields)
        c.execute("""DELETE FROM %s WHERE %s""" \
                        % (self.ref, where), wdata)

        if c.execute("""SELECT * FROM %s WHERE %s=%%s""" \
                            % (self.ref, self.r_field), cref) == 0:
            c.execute("""DELETE FROM %s WHERE %s=%%s"""\
                            % (self.table, self.t_field), cref)
        c.close()

        # remove local entry
        del self.data[cref]

class MythDBConn( object ):
    """
    Basic connection to the mythtv database
    Offers connection caching to prevent multiple connections
    """
    db = None

    class _TableFields(object):
        """Provides a dictionary-like list of table fieldnames"""
        def __init__(self, ident): self._ident = ident
        def __getattr__(self,key):
            if key in self.__dict__:
                return self.__dict__[key]
            elif key in MYTHDBCONN[self._ident][2]:
                return MYTHDBCONN[self._ident][2][key]
            else:
                return self.__getitem__(key)
        def __getitem__(self,key):
            if key in MYTHDBCONN[self._ident][2]:
                return MYTHDBCONN[self._ident][2][key]
            table_names = []
            c = MYTHDBCONN[self._ident][0].cursor()
            try:
                c.execute("DESC %s" % (key,))
            except:
                return None
            
            for name in c.fetchall():
                table_names.append(name[0])
            c.close()
            MYTHDBCONN[self._ident][2][key] = tuple(table_names)
            return table_names

    class _Settings(object):
        """Provides dictionary-like list of hosts"""
        class __Settings(object):
            """Provides dictionary-like list of settings"""
            def __init__(self,ident,host):
                self.data = {}
                self._ident = ident
                self._host = host
            def __getattr__(self,key):
                if key in self.__dict__:
                    return self.__dict__[key]
                elif key in self.data:
                    return self.data[key]
                else:
                    return self.__getitem__(key)
            def __getitem__(self,key):
                if key in self.data:
                    return self.data[key]
                if self._host == 'NULL':
                    where = 'IS NULL'
                    wheredat = (key,)
                else:
                    where = 'LIKE(%s)'
                    wheredat = (key, self._host)
                c = MYTHDBCONN[self._ident][0].cursor()
                c.execute("""SELECT data FROM settings WHERE value LIKE(%%s)
                    AND hostname %s LIMIT 1""" % where, wheredat)
                row = c.fetchone()
                c.close()
                if row:
                    self.data[key] = row[0]
                    return row[0]
                else:
                    return None
        def __init__(self, ident): self._ident = ident
        def __getattr__(self,key):
            if key in self.__dict__:
                return self.__dict__[key]
            elif key in MYTHDBCONN[self._ident][3]:
                return MYTHDBCONN[self._ident][3][key]
            else:
                return self.__getitem__(key)
        def __getitem__(self,key):
            if key not in MYTHDBCONN[self._ident][3]:
                MYTHDBCONN[self._ident][3][key] =\
                             self.__Settings(self._ident,key)
            return MYTHDBCONN[self._ident][3][key]

    def __init__(self, args=None):
        if args is not None:
            if isinstance(args, dict):
                # use connection settings from dictionary
                self.dbconn = args
            else:
                # load existing database connection
                self.dbconn = args.dbconn
            if not self._check_dbconn(self.dbconn):
                raise MythError("Required information for database "+\
                                            "connection missing")
        else:
            # try to read from config.xml
            config_files = [ os.path.expanduser('~/.mythtv/config.xml') ]
            if 'MYTHCONFDIR' in os.environ:
                config_files.append('%s/config.xml' %os.environ['MYTHCONFDIR'])

            dbconn = {'DBHostName':0,   'DBName':0, 'DBUserName':0,
                      'DBPassword':0,   'USN':0,    'SecurityPin':0}

            for config_file in config_files:
                try:
                    config = minidom.parse(config_file)
                except:
                    continue

                for key in dbconn:
                    dbconn[key] = None

                for token in config.getElementsByTagName('Configuration')[0].\
                                getElementsByTagName('UPnP')[0].\
                                getElementsByTagName('MythFrontend')[0].\
                                getElementsByTagName('DefaultBackend')[0].\
                                childNodes:
                    if token.nodeType == token.TEXT_NODE:
                        continue
                    try:
                        if token.tagName in dbconn:
                            dbconn[token.tagName] = token.childNodes[0].data
                    except:
                        pass

                if self._check_dbconn(dbconn):
                    #log.Msg(INFO, 'Using config %s', config_file)
                    self.dbconn = dbconn
                    break
            else:
                raise MythError("Could not find database login settings")

        self.ident = "sql://%s@%s/" % \
                (self.dbconn['DBName'],self.dbconn['DBHostName'])
        if self.ident in MYTHDBCONN:
            # reuse existing database connection and update count
            self.db = MYTHDBCONN[self.ident][0]
            MYTHDBCONN[self.ident][1] += 1
        else:
            # attempt new connection
            try:
                self.db = MySQLdb.connect(  user=  self.dbconn['DBUserName'],
                                            host=  self.dbconn['DBHostName'],
                                            passwd=self.dbconn['DBPassword'],
                                            db=    self.dbconn['DBName'],
                                            use_unicode=True,
                                            charset='utf8')
                #log.Msg(INFO, 'DB Connection info (host:%s, name:%s,'+
                #            ' user:%s, pass:%s)',   self.dbconn['DBHostName'],
                #                                    self.dbconn['DBName'],
                #                                    self.dbconn['DBUserName'],
                #                                    self.dbconn['DBPassword'])
            except:
                raise MythError("Connection failed for '%s'@'%s' to database "
                         "%s using password %s" % ( self.dbconn['DBUserName'],
                                                    self.dbconn['DBHostName'],
                                                    self.dbconn['DBName'],
                                                    self.dbconn['DBPassword']))
            # check schema version
            sver = self._check_schema('DBSchemaVer',SCHEMA_VERSION)
            if sver:
                #log.Msg(CRITICAL, 'DB speaks schema version %d, '+
                #                'but we speak version %d',
                #                sver, SCHEMA_VERSION)
                self.db.close()
                self.db = None
                raise MythError('Mismatched schema version')

            # add connection to cache
            MYTHDBCONN[self.ident] = [self.db, 1, {}, {}]

        # connect to table name cache
        self.tablefields = self._TableFields(self.ident)
        self.settings = self._Settings(self.ident)

    def _check_dbconn(self,dbconn):
        if      dbconn['DBHostName'] is None or \
                dbconn['DBName']     is None or \
                dbconn['DBUserName'] is None or \
                dbconn['DBPassword'] is None:
            return False
        else:
            return True

    def _check_schema(self,value,local):
        c = self.db.cursor()
        c.execute("""SELECT data FROM settings WHERE value LIKE(%s)""",(value,))
        sver = int(c.fetchone()[0])
        c.close()

        if local == sver:
            return 0
        else:
            return sver

    def getStorageGroup(self, groupname=None, hostname=None):
        """
        Returns tuple of dictionaries containing storage group directories
            with the fields 'id', 'groupname', 'hostname', and 'dirname'
        Takes an optional groupname and hostname for filtering
        """
        c = self.db.cursor()
        where = []
        wheredat = []
        if groupname:
            where.append("groupname=%s")
            wheredat.append(groupname)
        if hostname:
            where.append("hostname=%s")
            wheredat.append(hostname)
        if len(where):
            where = 'WHERE '+' AND '.join(where)
            c.execute("""SELECT * FROM storagegroup %s ORDER BY id"""\
                        % where, wheredat)
        else:
            c.execute("""SELECT * FROM storagegroup ORDER BY id""")

        ret = []
        for row in c.fetchall():
            ret.append(StorageGroup(row,self))
        return tuple(ret)

    def cursor(self): return self.db.cursor()

    def close(self): self.__del__()

    def __del__(self):
        if self.db is not None:
            # decrement use count
            MYTHDBCONN[self.ident][1] -= 1
            self.db = None

            if not MYTHDBCONN[self.ident][1]:
                # database connection no longer in use, close and delete it
                MYTHDBCONN[self.ident][0].close()
                del MYTHDBCONN[self.ident]


class MythBEConn( object ):
    """
    Basic class for mythbackend socket connections
    """

    db = None

    def __init__(self, backend=None, type='Monitor'):
        self.db = MythDBConn()
        if backend is None:
            # use master backend, no sanity checks, these should always be set
            self.host = self.db.settings.NULL.MasterServerIP
            self.port = self.db.settings.NULL.MasterServerPort
        else:
            if re.match('(?:\d{1,3}\.){3}\d{1,3}',backend):
                # process ip address
                c = self.db.cursor()
                if c.execute("""SELECT hostname FROM settings WHERE
                        value='BackendServerIP' AND
                        data=%s""", backend) == 0:
                    c.close()
                    raise MythError('Backend not found in database: '+backend)
                backend = c.fetchone()[0]
                c.close()
            self.host = self.db.settings[backend].BackendServerIP
            if not self.host:
                    raise MythError('Backend not found in database: '+backend)
            self.port = self.db.settings[backend].BackendServerPort
            if not self.port:
                raise MythError('Backend port not found in database: '+backend)
        self.port = int(self.port)

        try:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.settimeout(10)
            self.socket.connect((self.host, self.port))
            self.check_version()
            self.announce(type)
        except socket.error, e:
            #log.Msg(CRITICAL, 'Couldn\'t connect to %s:%d '+
            #            '(is the backend running)', self.host, self.port)
            raise MythError("Couldn't connect to %s:%d "+
                        "(is the backend running)" % (self.host,self.port))
        except:
            raise

    def __del__(self):
        self.backendCommand('DONE')
        self.socket.shutdown(1)
        self.socket.close()

    def close(self): self.__del__()

    def check_version(self):
        res = self.backendCommand('MYTH_PROTO_VERSION %s' \
                    % PROTO_VERSION).split(BACKEND_SEP)
        if res[0] == 'REJECT':
            #log.Msg(CRITICAL, 'Backend has version %s '+
            #        'and we speak version %s', res[1], PROTO_VERSION)
            raise MythError('Backend has version %s '+
                    'and we speak version %s' % (res[1], PROTO_VERSION))

    def announce(self,type):
        res = self.backendCommand('ANN %s %s 0' % (type, socket.gethostname()))
        if res != 'OK':
            #log.Msg(CRITICAL, 'Unexpected answer to ANN command: %s', res)
            raise MythError('Unexpected answer to ANN command: %s' % res)
        #else:
            #log.Msg(INFO, 'Successfully connected mythbackend at %s:%d',
            #                self.host, self.port)

    def backendCommand(self, data):
        """
        Sends a formatted command via a socket to the mythbackend.

        Returns the result from the backend.
        """
        def recv():
            """
            Reads the data returned from the backend.
            """
            # The first 8 bytes of the response gives us the length
            data = self.socket.recv(8)
            try:
                length = int(data)
            except:
                return u''
            data = []
            while length > 0:
                chunk = self.socket.recv(length)
                length = length - len(chunk)
                data.append(chunk)
            try:
                return unicode(''.join(data),'utf8')
            except:
                return u''.join(data)

        command = u'%-8d%s' % (len(data), data)
        #log.Msg(DEBUG, 'Sending command: %s', command)
        self.socket.send(command)
        return recv()

    def joinInt(self,high,low):
        """
        Returns a single long from a pair of signed integers
        """
        return (high + (low<0))*2**32 + low

    def splitInt(self,integer):
        """
        Returns a pair of signed integers from a single long
        """
        return integer/(2**32),integer%2**32 - (integer%2**32 > 2**31)*2**32



class MythXMLConn( object ):
    def __init__(self, backend=None):
        db = MythDBConn()
        if backend is None:
            # use master backend
            self.host = db.settings.NULL.MasterServerIP
            self.port = 6544 #self.db.settings.NULL.MasterServerPort
        elif re.match('(?:\d{1,3}\.){3}\d{1,3}',backend):
            # process ip address
            self.host = backend
            c = db.cursor()
            if c.execute("""SELECT hostname FROM settings WHERE
                    value='BackendServerIP' AND
                    data=%s""", self.host) == 0:
                backend = None
            else:
                backend = c.fetchone()[0]
            self.port = 6544 #self.db.settings[backend].BackendServerPort
            c.close()
        else:
            # assume given a hostname
            self.host = db.settings[backend].BackendServerIP
            if not self.host:
                # try a truncated hostname
                backend = backend.split('.')[0]
                self.host = db.setting[backend].BackendServerIP
                if not self.host:
                    raise MythError('Unable to find requested backend.')
            self.port = 6544 #self.db.setting[backend].BackendServerPort
        db.close()

    def _query(self, path=None, **keyvars):
        url = 'http://%s:%d/' % (self.host, self.port)
        if path is not None:
            url += 'Myth/%s/' % path
        if len(keyvars) > 0:
            fields = []
            for key in keyvars:
                fields.append('%s=%s' % (key, keyvars[key]))
            url += '?'+'&'.join(fields)
        print 'opening %s' % url
        ufd = urlopen(url)
        res = minidom.parseString(ufd.read())
        ufd.close()
        return res
    

class MythLog( object ):
    """
    A simple logging class
    """
    def __init__(self, level, format, instance):
        self.log = logging.getLogger(instance)
        self.log.setLevel(level)
        self.ch = logging.StreamHandler()
        self.ch.setFormatter(logging.Formatter(format))
        self.log.addHandler(self.ch)

    def Msg(self, level, msg, *args, **kwargs):
        self.log.log(level, msg, *args, **kwargs)

class MythError(Exception):
    """
    A simple exception class
    """
    def __init__(self, message):
        self.message = message

    def __repr__(self):
        print ': ' + self.message

class StorageGroup(DBData):
    """
    Represents a single storage group entry
    """
    table = 'storagegroup'
    where = 'id=%s'
    
    def __str__(self):
        return u"%s@%s%s" % (self.groupname,self.hostname,self.dirname)

    def __repr__(self): return self.__str__().encode('utf-8')

    def __init__(self,data,db=None):
        DBData.__init__(self,data,db)
        self.wheredat = (self.id)
        if self.hostname == socket.gethostname():
            self.local = True
        elif os.access(self.dirname, os.F_OK):
            self.local = True
        else:
            self.local = False





#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Provides base classes for accessing MythTV
"""

from MythStatic import *
from MythBase import *
from MythData import *

import os, re, socket

class MythTV( MythBEConn ):
    """
    A connection to a MythTV backend.
    """

    locked_tuners = []

    def __del__(self):
        self.freeTuner()
        MythBEConn.__del__(self)

    def getPendingRecordings(self):
        """
        Returns a list of Program objects which are scheduled to be recorded.
        """
        programs = []
        res = self.backendCommand('QUERY_GETALLPENDING').split(BACKEND_SEP)
        has_conflict = int(res.pop(0))
        num_progs = int(res.pop(0))
        #log.Msg(DEBUG, '%s pending recordings', num_progs)
        for i in range(num_progs):
            programs.append(Program(res[i * PROGRAM_FIELDS:(i * PROGRAM_FIELDS)
                + PROGRAM_FIELDS]))
        return tuple(programs)

    def getScheduledRecordings(self):
        """
        Returns a list of Program objects which are scheduled to be recorded.
        """
        programs = []
        res = self.backendCommand('QUERY_GETALLSCHEDULED').split(BACKEND_SEP)
        num_progs = int(res.pop(0))
        #log.Msg(DEBUG, '%s scheduled recordings', num_progs)
        for i in range(num_progs):
            programs.append(Program(res[i * PROGRAM_FIELDS:(i * PROGRAM_FIELDS)
                + PROGRAM_FIELDS]))
        return tuple(programs)

    def getUpcomingRecordings(self):
        """
        Returns a list of Program objects which are scheduled to be recorded.

        Sorts the list by recording start time and only returns those with
        record status of WillRecord.
        """
        def sort_programs_by_starttime(x, y):
            if x.starttime > y.starttime:
                return 1
            elif x.starttime == y.starttime:
                return 0
            else:
                return -1
        programs = []
        res = self.getPendingRecordings()
        for p in res:
            if p.recstatus == RECSTATUS['WillRecord']:
                programs.append(p)
        programs.sort(sort_programs_by_starttime)
        return tuple(programs)

    def getRecorderList(self):
        """
        Returns a list of recorders, or an empty list if none.
        """
        recorders = []
        c = self.db.cursor()
        c.execute('SELECT cardid FROM capturecard')
        row = c.fetchone()
        while row is not None:
            recorders.append(int(row[0]))
            row = c.fetchone()
        c.close()
        return recorders

    def getFreeRecorderList(self):
        """
        Returns a list of free recorders, or an empty list if none.
        """
        res = self.backendCommand('GET_FREE_RECORDER_LIST').split(BACKEND_SEP)
        recorders = [int(d) for d in res]
        return recorders

    def lockTuner(self,id=None):
        """
        Request a tuner be locked from use, optionally specifying which tuner
        Returns a tuple of ID, video dev node, audio dev node, vbi dev node
        Returns an ID of -2 if tuner is locked
                         -1 if no tuner could be found
        """
        local = True
        cmd = 'LOCK_TUNER'
        if id is not None:
            cmd += ' %d' % id
            res = self.getRecorderDetails(id).hostname
            if res != socket.gethostname():
                local = False

        res = ''
        if local:
            res = self.backendCommand(cmd).split(BACKEND_SEP)
        else:
            myth = MythTV(res)
            res = myth.backendCommand(cmd).split(BACKEND_SEP)
            myth.close()
        res[0] = int(res[0])
        if res[0] > 0:
            self.locked_tuners.append(res[0])
        return tuple(res)


    def freeTuner(self,id=None):
        """
        Frees a requested tuner ID
        If no ID given, free all tuners listed as used by this class instance
        """
        def free(self,id):
            res = self.getRecorderDetails(id).hostname
            if res == socket.gethostname():
                self.backendCommand('FREE_TUNER %d' % id)
            else:
                myth = MythTV(res)
                myth.backendCommand('FREE_TUNER %d' % id)
                myth.close()

        if id is None:
            for i in range(len(self.locked_tuners)):
                free(self,self.locked_tuners.pop())
        else:
            try:
                self.locked_tuners.remove(id)
            except:
                pass
            free(self,id)

    def getCurrentRecording(self, recorder):
        """
        Returns a Program object for the current recorders recording.
        """
        res = self.backendCommand('QUERY_RECORDER '+
                BACKEND_SEP.join([str(recorder),'GET_CURRENT_RECORDING']))
        return Program(res.split(BACKEND_SEP))

    def isRecording(self, recorder):
        """
        Returns a boolean as to whether the given recorder is recording.
        """
        res = self.backendCommand('QUERY_RECORDER '+
                BACKEND_SEP.join([str(recorder),'IS_RECORDING']))
        if res == '1':
            return True
        else:
            return False

    def isActiveBackend(self, hostname):
        """
        Returns a boolean as to whether the given host is an active backend
        """
        res = self.backendCommand(BACKEND_SEP.join(\
                    ['QUERY_IS_ACTIVE_BACKEND',hostname]))
        if res == 'TRUE':
            return True
        else:
            return False

    def getRecording(self, chanid, starttime):
        """
        Returns a Program object matching the channel id and start time
        """
        res = self.backendCommand('QUERY_RECORDING TIMESLOT %d %d' \
                        % (chanid, starttime)).split(BACKEND_SEP)
        if res[0] == 'ERROR':
            return None
        else:
            return Program(res[1:])

    def getRecordings(self):
        """
        Returns a list of all Program objects which have already recorded
        """
        programs = []
        res = self.backendCommand('QUERY_RECORDINGS Play').split(BACKEND_SEP)
        num_progs = int(res.pop(0))
        #log.Msg(DEBUG, '%s total recordings', num_progs)
        for i in range(num_progs):
            programs.append(Program(res[i * PROGRAM_FIELDS:(i*PROGRAM_FIELDS)
                + PROGRAM_FIELDS]))
        return tuple(programs)

    def getExpiring(self):
        """
        Returns a tuple of all Program objects nearing expiration
        """
        programs = []
        res = self.backendCommand('QUERY_GETEXPIRING').split(BACKEND_SEP)
        num_progs = int(res.pop(0))
        for i in range(num_progs):
            programs.append(Program(res[i * PROGRAM_FIELDS:(i*PROGRAM_FIELDS)
                + PROGRAM_FIELDS]))
        return tuple(programs)

    def getCheckfile(self,program):
        """
        Returns location of recording in file system
        """
        res = self.backendCommand(BACKEND_SEP.join(\
                        ['QUERY_CHECKFILE','1',program.toString()]\
                    )).split(BACKEND_SEP)
        if res[0] == 0:
            return None
        else:
            return res[1]

    def deleteRecording(self,program,force=False):
        """
        Deletes recording, set 'force' true if file cannot be found
        Returns the number fewer recordings that exist afterwards
        """
        command = 'DELETE_RECORDING'
        if force:
            command = 'FORCE_DELETE_RECORDING'
        return self.backendCommand(BACKEND_SEP.join(\
                    [command,program.toString()]))

    def forgetRecording(self,program):
        """
        Forgets old recording and allows it to be re-recorded
        """
        self.backendCommand(BACKEND_SEP.join(['FORGET_RECORDING',
                    program.toString()]))

    def deleteFile(self,file,sgroup):
        """
        Deletes a file from specified storage group on the connected backend
        Takes a relative file path from the root of the storage group
        Returns 1 on success
        """
        return self.backendCommand(BACKEND_SEP.join(\
                    ['DELETE_FILE',file,sgroup]))

    def getFreeSpace(self,all=False):
        """
        Returns a tuple of FreeSpace objects
        """
        command = 'QUERY_FREE_SPACE'
        if all:
            command = 'QUERY_FREE_SPACE_LIST'
        res = self.backendCommand(command).split(BACKEND_SEP)
        dirs = []
        for i in range(0,len(res)/10):
            dirs.append(res[i*10:i*10+18])
        return tuple(dirs)

    def getFreeSpaceSummary(self):
        """
        Returns a tuple of total space (in KB) and used space (in KB)
        """
        res = self.backendCommand('QUERY_FREE_SPACE_SUMMARY')\
                    .split(BACKEND_SEP)
        return (self.joinInt(int(res[0]),int(res[1])),
                self.joinInt(int(res[2]),int(res[3])))

    def getLoad(self):
        """
        Returns a tuple of the 1, 5, and 15 minute load averages
        """
        res = self.backendCommand('QUERY_LOAD').split(BACKEND_SEP)
        return (float(res[0]),float(res[1]),float(res[2]))

    def getUptime(self):
        """
        Returns machine uptime in seconds
        """
        return self.backendCommand('QUERY_UPTIME')

    def getSGList(self,host,sg,path,filenamesonly=False):
        """
        Returns a tuple of directories, files, and filesizes.
        Two special modes
            'filenamesonly' returns only filenames, no directories or size
            empty path base directories returns base storage group paths
        """
        if filenamesonly: path = '' # force empty path
        res = self.backendCommand(BACKEND_SEP.join(\
                    ['QUERY_SG_GETFILELIST',host,sg,path,
                    str(int(filenamesonly))]\
                )).split(BACKEND_SEP)
        if res[0] == 'EMPTY LIST':
            return -1
        if res[0] == 'SLAVE UNREACHABLE: ':
            return -2
        dirs = []
        files = []
        sizes = []
        for entry in res:
            if filenamesonly:
                files.append(entry)
            elif path == '':
                type,name = entry.split('::')
                dirs.append(name)
            else:
                type,name,size = entry.split('::')
                if type == 'file':
                    files.append(name)
                    sizes.append(size)
                if type == 'dir':
                    dirs.append(name)
        if filenamesonly:
            return tuple(files)
        elif path == '':
            return tuple(dirs)
        else:
            return (tuple(dirs),tuple(files),tuple(sizes))

    def getSGFile(self,host,sg,path):
        """
        Returns a tuple of last modification time and file size
        """
        res = self.backendCommand(BACKEND_SEP.join(\
                ['QUERY_SG_FILEQUERY',host,sg,path])).split(BACKEND_SEP)
        if res[0] == 'EMPTY LIST':
            return -1
        if res[0] == 'SLAVE UNREACHABLE: ':
            return -2
        return tuple(res[1:3])

    def getLastGuideData(self):
        """
        Returns the last dat for which guide data is available
        On error, 0000-00-00 00:00 is returned
        """
        return self.backendCommand('QUERY_GUIDEDATATHROUGH')


class Frontend(object):
    isConnected = False
    socket = None
    host = None
    port = None

    def __init__(self, host, port):
        self.host = host
        self.port = int(port)
        self.connect()
        self.disconnect()

    def __del__(self):
        if self.isConnected:
            self.disconnect()

    def __repr__(self):
        return "%s@%d" % (self.host, self.port)

    def __str__(self):
        return "%s@%d" % (self.host, self.port)

    def close(self): self.__del__()

    def connect(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(10)
        self.socket.connect((self.host, self.port))
        if re.search("MythFrontend Network Control.*",self.recv()) is None:
            self.socket.close()
            self.socket = None
            raise Exception('FrontendConnect',
                'Connected socket does not belong to a mythfrontend')
        self.isConnected = True

    def disconnect(self):
        self.send("exit")
        self.socket.close()
        self.socket = None
        self.isConnected = False

    def send(self,command):
        if not self.isConnected:
            self.connect()
        self.socket.send("%s\n" % command)

    def recv(self):
        curstr = ''
        prompt = re.compile('([\r\n.]*)\r\n# ')
        while not prompt.search(curstr):
            try:
                curstr += self.socket.recv(100)
            except socket.error:
                raise MythError('Frontend has closed connection')
            except KeyboardInterrupt:
                raise
        return prompt.split(curstr)[0]

    def sendJump(self,jumppoint):
        """
        Sends jumppoint to frontend
        """
        self.send("jump %s" % jumppoint)
        if self.recv() == 'OK':
            return 0
        else:
            return 1

    def getJump(self):
        """
        Returns a tuple containing available jumppoints
        """
        self.send("help jump")
        return tuple(re.findall('(\w+)[ ]+- ([\w /,]+)',self.recv()))

    def sendKey(self,key):
        """
        Sends keycode to connected frontend
        """
        self.send("key %s" % key)
        if self.recv() == 'OK':
            return 0
        else:
            return 1

    def getKey(self):
        """
        Returns a tuple containing available special keys
        """
        self.send("help key")
        return tuple(self.recv().split('\r\n')[4].split(', '))

    def sendQuery(self,query):
        """
        Returns query from connected frontend
        """
        self.send("query %s" % query)
        return self.recv()

    def getQuery(self):
        """
        Returns a tuple containing available queries
        """
        self.send("help query")
        return tuple(re.findall('query ([\w ]*\w+)[ \r\n]+- ([\w /,]+)',
                        self.recv()))

    def sendPlay(self,play):
        """
        Send playback command to connected frontend
        """
        self.send("play %s" % play)
        if self.recv() == 'OK':
            return 0
        else:
            return 1

    def getPlay(self):
        """
        Returns a tuple containing available playback commands
        """
        self.send("help play")
        return tuple(re.findall('play ([\w -:]*\w+)[ \r\n]+- ([\w /:,\(\)]+)',
                        self.recv()))


class MythDB( MythDBConn ):
    """
    Several canned queries for general purpose access to the database
    """
    settings = {}

    def getAllSettings(self, hostname=None):
        """
        Returns values for all settings.

        Returns None if there are no settings. If multiple rows are
        found (multiple hostnames), returns the value of the first one.
        """
        #log.Msg(DEBUG, 'Retrieving all setting for host %s', hostname)
        c = self.db.cursor()
        if hostname is None:
            hostname = 'NULL'
            c.execute("""
                    SELECT value, data
                    FROM settings
                    WHERE hostname IS NULL""")
        else:
            c.execute("""
                    SELECT value, data
                    FROM settings
                    WHERE hostname LIKE(%s)""",
                    (hostname))
        rows = c.fetchall()
        c.close()

        if rows:
            for row in rows:
                if str(hostname) not in self.settings:
                    self.settings[hostname] = {}
                if row[0] not in self.settings[str(hostname)]:
                    self.settings[hostname][row[0]] = row[1]
            return rows
        else:
            return None

    def searchRecorded(self, title=None, subtitle=None, chanid=None,
                        starttime=None, category=None, hostname=None,
                        autoexpire=None, commflagged=None, stars=None,
                        recgroup=None, playgroup=None, duplicate=None,
                        transcoded=None, watched=None, storagegroup=None,
                        cast=None):
        """
        Tries to find recordings matching the given information.
        Returns a tuple of Recorded objects
        """

        where = []
        fields = []
        tables = ['recorded']
        if title:
            where.append('recorded.title=%s')
            fields.append(title)
        if subtitle:
            where.append('recorded.subtitle=%s')
            fields.append(subtitle)
        if chanid:
            where.append('recorded.chanid=%s')
            fields.append(int(chanid))
        if starttime:
            where.append('recorded.starttime=%s')
            fields.append(starttime)
        if category:
            where.append('recorded.category=%s')
            fields.append(category)
        if hostname:
            where.append('recorded.hostname=%s')
            fields.append(host)
        if autoexpire:
            where.append('recorded.autoexpire=%s')
            fields.append(autoexpire)
        if commflagged:
            where.append('recorded.commflagged=%s')
            fields.append(commflagged)
        if stars:
            where.append('recorded.stars=%s')
            fields.append(stars)
        if recgroup:
            where.append('recorded.recgroup=%s')
            fields.append(recgroup)
        if playgroup:
            where.append('recorded.playgroup=%s')
            fields.append(playgroup)
        if duplicate:
            where.append('recorded.duplicate=%s')
            fields.append(duplicate)
        if transcoded:
            where.append('recorded.transcoded=%s')
            fields.append(transcoded)
        if watched:
            where.append('recorded.watched=%s')
            fields.append(watched)
        if storagegroup:
            where.append('recorded.storagegroup=%s')
            fields.append(storagegroup)
        if cast:
            tables.append('recordedcredits')
            where.append('recordedcredits.chanid=recorded.chanid')
            where.append('recordedcredits.starttime=recorded.starttime')
            fields.append(cast)
            if isinstance(cast,int):
                where.append('videometadatacast.person=%s')
            else:
                tables.append('people')
                where.append('recordedcredits.person=people.person')
                where.append('people.name=%s')

        if len(where) == 0:
            return None
        where  = ' AND '.join(where)
        tables = ','.join(tables)

        c = self.cursor()
        c.execute("""SELECT recorded.* FROM %s
                    WHERE %s""" % (tables,where), fields)
        rows = c.fetchall()
        c.close()

        if len(rows) == 0:
            return None
        records = []
        for row in rows:
            records.append(Recorded(data=row,db=self))
        if len(records):
            return tuple(records)
        else:
            return None

    def getFrontends(self):
        """
        Returns a list of Frontend objects for accessible frontends
        """
        cursor = self.db.cursor()
        cursor.execute("""  SELECT DISTINCT hostname
                            FROM settings
                            WHERE hostname IS NOT NULL
                            AND value='NetworkControlEnabled'
                            AND data=1""")
        frontends = []
        for fehost in cursor.fetchall():
            try:
                frontend = self.getFrontend(fehost[0])
                frontends.append(frontend)
            except:
                print "%s is not a valid frontend" % fehost[0]
        cursor.close()
        return frontends

    def getFrontend(self,host):
        """
        Returns a Frontend object for the specified host
        """
        port = self.db.setting[host].NetworkControlPort
        return Frontend(host,port)

    def getRecorded(self, title=None, subtitle=None, chanid=None,
                        starttime=None):
        """
        Tries to find a recording matching the given information.
        Returns a Recorded object.
        """
        records = self.searchRecorded(title,subtitle,season,\
                                episode,host,file,exactfile)
        if records:
            return records[0]
        else:
            return None

    def getChannels(self):
        """
        Returns a tuple of channel object defined in the database
        """
        channels = []
        c = self.db.cursor()
        c.execute("""SELECT * FROM channel""")
        for row in c.fetchall():
            channels.append(Channel(row,self))
        c.close()
        return tuple(channels)

    def getChannel(self,chanid):
        """
        Returns a single channel object for the given chanid
        """
        c = self.db.cursor()
        if c.execute("""SELECT * FROM channel
                WHERE chanid=%s""", (chanid,)):
            return Channel(c.fetchone(),self)
        else:
            return None

    def getGuideData(self, chanid, date):
        """
        Returns tuple of guide data for one channel on one date.
        Date is in the format YYYYMMDD.
        """
        guide = []
        c = self.db.cursor()
        c.execute("""SELECT * FROM program
                WHERE chanid=%s AND DATE(starttime)=%s""",
                (chanid,date))
        for show in c.fetchall():
            guide.append(Guide(show,self))
        c.close()
        return tuple(guide)


class MythVideo( MythDBConn ):
    """
    Provides convenient methods to access the MythTV MythVideo database.
    """
    def __init__(self,db=None):
        """
        Initialise the MythDB connection.
        """
        MythDBConn.__init__(self,db)

        # check schema version
        sver = self._check_schema('mythvideo.DBSchemaVer',MVSCHEMA_VERSION)
        if sver:
            #log.Msg(CRITICAL,
            #        'DB speaks schema version %d, but we speak version %d',
            #        sver, MVSCHEMA_VERSION)
            self.close()
            raise MythError('Mismatched schema version')

    def pruneMetadata(self):
        """
        Removes metadata from the database for files that no longer exist.
        Respects 'Videos' storage groups and relative paths.
        """
        c = self.cursor()
        c.execute("""SELECT DISTINCT host FROM videometadata""")
        hosts = c.fetchall()
        c.close()

        for h in hosts:
            host = h[0]
            myth = None
            if host is None:
                if self.getVideoSG(local=True) is None:
                    continue    # no local video directories to process
            else:
                try:
                    myth = MythTV(host)
                except MythError:
                    continue    # cannot connect to remote backend for scan

            videos = self.searchVideos(host=host)
            bases = self.getVideoSG(host)
            paths = {}
            for vid in videos:
                if host is None:    # check for local videos
                    if not os.path.exists(vid.filename):
                        vid.delete()
                    continue

                # scan available storage group paths
                vpath = '/'+vid.filename
                vfile = vpath[vpath.rfind('/')+1:]
                vpath = vpath[:vpath.rfind('/')]

                if vpath not in paths:
                    files = []
                    for b in bases:
                        d,f,s = myth.getSGList(host,'Videos',b+vpath)
                        files.extend(f)
                    paths[vpath] = files
                if vfile not in paths[vpath]:
                    print u'deleting '+str(vid)
                    vid.delete()


    def searchVideos(self, title=None, subtitle=None, season=None,
                        episode=None, host=None, file=None, exactfile=False,
                        cast=None, genre=None, country=None, category=None):
        """
        Tries to find videos matching the given information.
        Returns a tuple of Video objects
        """
        if host:
            if self.getVideoSG(host):
                host = host.lower()
            else:
                host = None
        where = []
        fields = []
        tables = ['videometadata']
        if title:
            where.append('videometadata.title=%s')
            fields.append(title)
        if subtitle:
            where.append('videometadata.subtitle=%s')
            fields.append(subtitle)
        if season:
            where.append('videometadata.season=%s')
            fields.append(int(season))
        if episode:
            where.append('videometadata.episode=%s')
            fields.append(int(episode))
        if host:
            where.append('videometadata.host=%s')
            fields.append(host)
        if file:
            if exactfile:
                where.append('videometadata.filename=%s')
                fields.append(file)
            else:
                where.append('videometadata.filename LIKE %s')
                fields.append('%'+file)
        if cast:
            tables.append('videometadatacast')
            where.append('videometadata.intid=videometadatacast.idvideo')
            fields.append(cast)
            if isinstance(cast,int):
                where.append('videometadatacast.idcast=%s')
            else:
                tables.append('videocast')
                where.append('videometadatacast.idcast=videocast.intid')
                where.append('videocast.cast=%s')
        if genre:
            tables.append('videometadatagenre')
            where.append('videometadata.intid=videometadatagenre.idvideo')
            fields.append(genre)
            if isinstance(genre,int):
                where.append('videometadatagenre.idgenre=%s')
            else:
                tables.append('videogenre')
                where.append('videometadatagenre.idgenre=videogenre.intid')
                where.append('videogenre.genre=%s')
        if country:
            tables.append('videometadatacountry')
            where.append('videometadata.intid=videometadatacountry.idvideo')
            fields.append(country)
            if isinstance(country,int):
                where.append('videometadatacountry.idcountry=%s')
            else:
                tables.append('videocontry')
                where.append('videometadatacountry.idcountry=videocountry.intid')
                where.append('videocast.country=%s')
        if category:
            fields.append(category)
            if isinstance(category,int):
                where.append('vidoemetadata.category=%s')
            else:
                tables.append('videocategory')
                where.append('videometadata.category=videocategory.intid')
                where.append('videocategory.category=%s')

        if len(where) == 0:
            return None
        where  = ' AND '.join(where)
        tables = ','.join(tables)

        c = self.cursor()
        c.execute("""SELECT videometadata.* FROM %s
                    WHERE %s""" % (tables,where), fields)
        rows = c.fetchall()
        c.close()

        if len(rows) == 0:
            return None
        videos = []
        for row in rows:
            videos.append(Video(data=row,db=self))
        if len(videos):
            return tuple(videos)
        else:
            return None

    def getVideo(self, title=None, subtitle=None, season=None,
                        episode=None, host=None, file=None, exactfile=False):
        """
        Tries to find a video matching the given information.
        Returns a Video object.
        """
        videos = self.searchVideos(title,subtitle,season,\
                                episode,host,file,exactfile)
        if videos:
            return videos[0]
        else:
            return None

